@@ -1,9 +1,8 @@
 #!/usr/bin/env python3
 
 import asyncio
-import logging
 from collections import deque
-from pprint import pprint  # noqa: W0611
+from pprint import pprint  # noqa: W0611 F401
 from threading import Barrier, Thread
 
 import janus
@@ -13,8 +12,6 @@ from .component import Component
 from .pipe import Manifold, Message
 from .utils import aiter, isdict, islist
 
-logger = logging.getLogger(__name__)
-
 
 class Node(Component):
     """Processes messages."""
@@ -30,6 +27,7 @@ class Node(Component):
         self.ofilters = ofilters or PriorityRegistry()
         self.outbox = self._join_pipes(outbox, Outbox)
         self.layers = list()
+        self.streams = dict()
         return settings
 
     def __str__(self):
@@ -58,6 +56,14 @@ class Node(Component):
                 coroutines.append(layer.start())
         return asyncio.gather(*coroutines)
 
+    async def test(self, messages):
+        self.debug(f'~~~~~~ begin {messages}')
+        async for c, m in messages:
+            self.debug(f'~~~~~~ message from {messages} {c} {m}')
+            yield c, m
+            self.debug(f'~~~~~~ yielded from {messages} {c} {m}')
+        self.debug(f'~~~~~~ end {messages}')
+
     async def run(self):
         await super().run()
         for layer in self.layers:
@@ -70,15 +76,17 @@ class Node(Component):
             stack = self.layers[0]
             for layer in self.layers[1:]:
                 stack = layer(stack)
+                # stack = layer(self.test(stack))
             self.debug('Processing messages')
             try:
                 async for result in stack:
+                # async for result in self.test(stack):
                     result
             except Exception:
-                self.exception(f'Node failure')
+                self.exception('Node failure')
                 raise
 
-            self.debug(f'Finished processing')
+            self.debug('Finished processing')
 
     def spawn_processor(self):
         processor = self.get_processor()
@@ -215,14 +223,33 @@ class Inbox(Manifold):
         raise NotImplementedError(f'receive() in {self}')
 
     async def receiver(self):
+        message = None
+        active_channels = 0
+        last_channel = False
         self.debug('Receiving')
         for channel, pipe in self._active_channels():
             if not self.running:
                 return
 
+            if pipe is None:
+                if active_channels == 1:
+                    last_channel = True
+                if message is False:  # all channels are empty
+                    # self.debug(f'***** channels sweep empty')
+                    await asyncio.sleep(0.05)
+                    # self.debug(f'***** resuming channels sweep')
+
+                message = False
+                active_channels = 0
+                continue
+
+            active_channels += 1
             try:
-                # self.debug(f'* waiting for message on channel {channel}')
-                message = await pipe.receive()
+                # self.debug(f'***** waiting for message on channel {channel}')
+                message = await pipe.receive(wait=last_channel)
+            except asyncio.QueueEmpty:
+                continue
+
             except Exception:
                 self.exception(f'Error while receiving on channel {channel}, {pipe}')
                 raise
@@ -240,6 +267,8 @@ class Inbox(Manifold):
     def _active_channels(self):
         active_channels = self.channels.copy()
         while active_channels:
+            yield None, None  # signal start of channels sweep
+
             for channel, pipe in active_channels.copy().items():
                 if not pipe.running:
                     self.debug(f'skipping stopped channel {channel}')
@@ -308,7 +337,7 @@ class Inbox(Manifold):
                 self.exception(f'Error while receiving on channel {channel}, {pipe}')
                 raise
 
-            # self.debug(f'* received from channel {channel} (next event: {events[channel].is_set()})')
+            # self.debug(f'* received from channel {channel} (next event: {events[channel].is_set()})')  # noqa: E501
             events.received.set()
         self.debug(f'Receiver for {channel} is exiting')
 
@@ -341,17 +370,17 @@ class RetryableMessages:
     async def feeder(self):
         """Feed the inbox to the queue."""
         async for channel, self.newmsg in self.messages:
-            # self.node.debug(f'*** collector retry: feeding {self.newmsg.type} msg {self.newmsg.meta.id}')
+            # self.node.debug(f'*** collector retry: feeding {self.newmsg.type} msg {self.newmsg.meta.id}')  # noqa: E501
             self.forwarded.clear()
             await self.queue.put((channel, self.newmsg))
             await self.forwarded.wait()
-        # self.node.debug(f'*** collector retry: feeder done')
+        # self.node.debug('*** collector retry: feeder done')
         self.exhausted = True
         if self.pending > 0:
             await self.finished.wait()
-        # self.node.debug(f'*** collector retry: feeder ready to exit')
+        # self.node.debug('*** collector retry: feeder ready to exit')
         await self.queue.put(Message.EOT)
-        # self.node.debug(f'*** collector retry: feeder exited')
+        # self.node.debug('*** collector retry: feeder exited')
 
     async def forwarder(self):
         """Turn the queue into an iterator that the processor can consume."""
@@ -363,7 +392,7 @@ class RetryableMessages:
         self.pending = 0  # number of messages currently handled by node processor
         feeder = self.node.loop.create_task(self.feeder())
         async for channel, message in aiter(self.queue.get, Message.EOT):
-            # self.node.debug(f'*** collector retry: forwarding {message.type} msg {message.meta.id}')
+            # self.node.debug(f'*** collector retry: forwarding {message.type} msg {message.meta.id}')  # noqa: E501
             self.queue.task_done()
             if message is self.newmsg:
                 self.forwarded.set()
@@ -384,7 +413,7 @@ class RetryableMessages:
 
     async def retry(self, channel, message, wait=0.001, id_=None):
         """Insert a message back into the queue for retrying it."""
-        # self.node.debug(f'*** collector retry{"" if id_ is None else " %s" % id_}: retrying {message.type} msg {message.meta.id}')
+        # self.node.debug(f'*** collector retry{"" if id_ is None else " %s" % id_}: retrying {message.type} msg {message.meta.id}')  # noqa: E501
         await asyncio.sleep(wait)
         await self.node.loop.create_task(self.queue.put((channel, message)))
         self.processing_done()
@@ -401,7 +430,7 @@ class Outbox(Manifold):
             if channel is Message.DROP:
                 self.drop(message)
             else:
-                self.debug(f'Sending to {channel}: {message}')
+                self.debug(f'Sending message to {channel}: {message}')
                 if channel in self.channels:
                     await self.channels[channel].send(message)
                     # self.debug(f'Sent to    {channel}: {message} [{self.channels[channel].id}]')
@@ -414,67 +443,11 @@ class Outbox(Manifold):
 
         self.debug('Finished sending')
         for pipe in self.channels.values():
-         try:
-            self.debug(f'EOT to {pipe}')
-            await pipe.send(Message.EOT)
-         except:
-            __import__('pudb').set_trace()
-
-
-class Filter(Component):
-
-    def __call__(self, messages):
-        return self.filter(messages)
-
-    async def filter(self, messages):
-        async for channel, message in messages:
-            channel, message = await self.process(channel, message)
-            yield (channel, message)
-
-            if not self.running:
-                break
-
-    async def process(self, channel, message):
-        raise NotImplementedError(f'process() out {self}')
-
-
-class ChannelChain(Filter):
-
-    TESTS = dict(directory=lambda m: m.path.id_dir(),
-                 file=lambda m: m.path.id_file())
-
-    async def process(self, channel, message):
-        for test_name, test_fn in self.TESTS.items():
-            if test_fn(message):
-                return test_name, message
-
-
-class MessageMangleFilter(Filter):
-
-    def deepget(self, mapping, key):
-        if '.' not in key:
-            return mapping[key]
-
-        else:
-            current, remainder = key.split('.', 1)
-            return self.deepget(mapping[current], remainder)
-
-    def deepset(self, mapping, key, value):
-        if '.' not in key:
-            mapping[key] = value
-
-        else:
-            current, remainder = key.split('.', 1)
-            self.deepset(mapping[current], remainder, value)
-
-    async def process(self, channel, message):
-        for spec in self.settings.attributes:
-            if len(spec) != 1:
-                raise ValueError(f'Bad configuration: {self.settings}')
-
-            key = list(spec.keys())[0]
-            self.deepset(message, key, spec[key].format(**message))
-        return channel, message
+            try:
+                self.debug(f'EOT to {pipe}')
+                await pipe.send(Message.EOT)
+            except Exception:
+                __import__('pudb').set_trace()
 
 
 class ThreadedNode(Node):
@@ -510,7 +483,7 @@ class ThreadedProcessor:
             # print(f'thread {i} EOT received back')
 
         while self.threads:  # FIXME: join threads instead, and move or retire del self.threads[]
-            asyncio.sleep(0.1)  #     while at it, catch exceptions and clean up queue with task_done  # noqa: E501
+            asyncio.sleep(0.1)      # while at it, catch exceptions and clean up queue with task_done  # noqa: E501
         # print(f'all threads stopped')
 
     async def thread_feeder(self, messages):
@@ -555,7 +528,7 @@ class CmdRunner:
             stderr = (await proc.stderr.read()).decode()
             if exitcode:
                 raise RuntimeError(f'Command {args} ({kwargs}) returned with exitcode {exitcode}, '
-                                f'stdout: {stdout or None}, stderr: {stderr or None}')
+                                   f'stdout: {stdout or None}, stderr: {stderr or None}')
 
         except Exception as e:
             self.logger.exception(f'Error running command: {e}')
